/** An example Reactor controller for creating thermostat-like schedules
 *  Copyright (c) 2024 Patrick H. Rigney, All Rights Reserved.
 *  WhiteBitController is offered under MIT License - https://mit-license.org/
 *  More info: https://reactor.toggledbits.com/
 */

const version = 24096;

const Controller = require("server/lib/Controller");

const Logger = require( "server/lib/Logger" );
Logger.getLogger( 'ScheduleController', 'Controller' ).always( "Module ScheduleController v%1", version );

const Configuration = require("server/lib/Configuration");
const logsdir = Configuration.getConfig( "reactor.logsdir" );  /* logs directory path if you need it */

/* Another other modules we need? */
const util = require("server/lib/util");

var impl = false;  /* Implementation data, one copy for all instances, will be loaded by start() later */

var toSSM = function( d ) {
    return d.getHours() * 3600 + d.getMinutes() * 60 + d.getSeconds()
};

module.exports = class ScheduleController extends Controller {
    constructor( struct, id, config ) {
        super( struct, id, config );  /* required */
    }

    /** Start the controller.
     */
    async start() {
        /* Find the sunrise/sunset entity */
        this.suninfo = this.getStructure().findEntity( this.config.sun_entity || "reactor_system>sun" );

        /* Make sure our configuration gives us data to query */
        for ( let eid in ( this.config.entities || {} ) ) {
            let c = this.config.entities[ eid ];
            let e = this.findEntity( eid );
            if ( ! e ) {
                e = this.getEntity( "ValueSensor", eid );
                e.extendCapability( "string_sensor" );
                e.setPrimaryAttribute( "string_sensor.value" );
                e.setName( c.name || eid );
            }
        }

        this.startDelay( 1 );

        this.online();

        /** Must return `this` or a Promise that resolves to this. If this
         *  is declared async (as it should be), the return of a Promise
         *  is implicit (async functions always return Promise).
         */
        return this;
    }

    /** Stop the controller.
     *  You only need to supply this method if your subclass has something
     *  specific it needs to do.
     */
    async stop() {
        this.log.notice( "%1 stopping", this );
        this.stopTimer();

        /* Required ending */
        return await super.stop();
    }

    /** run() is called when Controller's single-simple timer expires.
     *  If you don't use Controller's simple timer, you don't need this.
     */
    run() {
        const d = new Date();
        const now = toSSM( d );
        let nextEdge = null;
        for ( let eid in this.config.entities ) {
            this.log.debug( 5, "%1 handling entity %2", this, eid );
            let c = this.config.entities[ eid ];
            let edges = [];
            for ( let edge in ( c.schedule || {} ) ) {
                edges.push( { entity: eid, edge: edge, time: this._parse_edge( edge ), value: c.schedule[ edge ] } );
            }
            if ( edges.length < 1 ) {
                continue;
            }
            // Sort edges by establishes times for today.
            edges.sort( (a,b) => {
                return a.time - b.time;
            } );
            if ( edges[0].time > 0 ) {
                // First edge is not midnight, so create a midnight/wrap-around edge from the last edge.
                let last = { ...edges[edges.length-1] };
                last.edge = "00:00";
                last.time = 0; // midnight
                edges.unshift( last );
            }
            this.log.debug( 5, "%1 entity %2 edges %3", this, eid, edges );

            // Find the current period in local time.
            let last = edges[0];
            for ( let edge of edges ) {
                if ( edge.time > now ) {
                    if ( nextEdge === null || edge.time < nextEdge.time ) {
                        nextEdge = edge;
                    }
                    break;
                }
                last = edge;
            }

            let e = this.getEntity( "ValueSensor", eid );
            if ( e.getAttribute( 'string_sensor.value' ) !== last.value ) {
                this.log.info( "%1 entity %2 edge %3 active, setting primary attribute value to %4", this, e, last.edge, last.value );
                e.deferNotifies( true );
                try {
                    let since = new Date();
                    since.setHours( 0, 0, 0 );
                    since.setSeconds( last.time );  // last.time is SSM
                    e.setAttribute( 'string_sensor.value', last.value );
                    e.setAttribute( 'x_schedule_info.edge', last.edge );
                    e.setAttribute( 'x_schedule_info.since', since.getTime() );
                    e.setAttribute( 'x_schedule_info.asof', since.toLocaleTimeString() );
                    e.setAttribute( 'x_schedule_info.edges', JSON.stringify( c.schedule ) );
                } catch ( err ) {
                    this.log.exception( err );
                } finally {
                    e.deferNotifies( false );
                }
            }
        }
        
        // nextEdge, if non-null, is now the earliest next edge for all entities.
        if ( nextEdge ) {
            let nd = new Date();
            nd.setHours( Math.floor( nextEdge.time / 3600 ), Math.floor( nextEdge.time / 60 ) % 60, nextEdge.time % 60 );
            if ( nd <= d ) {
                // Next date/time computed before current, go to next day (next edge is at/across midnight).
                nd.setDate( nd.getDate() + 1 );
            }
            this.log.debug( 5, "%1 next event %2 %3 at %4", this, nextEdge.entity, nextEdge.edge, nd.toLocaleString() );

            // Find the delay to the next edge. Rather than just computing seconds straight-away, we DST-proof the
            // calculation by making the Date object compute the difference between the two dates.
            this.stopTimer();
            let delay = nd - Date.now();
            this.log.debug( 5, "%1 delay to next event %2 time %3", this, delay, new Date( Date.now() + delay ) );
            this.startDelay( delay );
            this.getSystemEntity().setAttributes({
                next_event:         [ nd.getTime(), nextEdge.entity, nextEdge.edge, nextEdge.value ],
                next_event_date:    nd.toLocaleString()
            }, "x_scheduler");
        }
    }

    _parse_edge( edge ) {
        let m = edge.match( /^\s*(\d+):(\d+)(:(\d+))?/ );
        if ( m ) {
            let moment = parseInt( m[1] ) * 3600 + parseInt( m[2] ) * 60 + util.coalesce( parseInt( m[4] ), 0 );
            return moment;
        }
        m = edge.toLowerCase().match( /^(sunrise|sunset)(\s*(\+|-)\s*(.*))?/ );
        if ( m ) {
            let ed = new Date( this.suninfo.getAttribute( `suninfo.${m[1]}` ) );
            let moment = toSSM( ed );
            this.log.debug( 5, "%1 edge %2 established %3 at %4 moment %5", this, edge, m[1], ed.toLocaleTimeString(), moment );
            if ( m[3] ) {  // +/-
                let offset = 0;
                let sign = m[3] === "-" ? -1 : 1;
                let rest = m[4] || "";
                this.log.debug( 5, "%1 edge %2 parsing offset %3", this, edge, rest );
                while ( rest.trim().length ) {
                    let p = rest.match( /^(\d+)\s*(h(ou)?rs?|min(ute)?s?|sec(ond)s?)\s*/ );
                    if ( p ) {
                        this.log.debug( 5, "%1 for %2 handling chunk %3", this, edge, p[0] );
                        let n = parseInt( p[1] );
                        if ( p[2].startsWith( 'h' ) ) {
                            n *= 3600;
                        } else if ( p[2].startsWith( 'm' ) ) {
                            n *= 60;
                        }
                        offset += sign * n;
                        rest = rest.substr( p[0].length );
                    } else {
                        p = rest.match( /\s*(\d+):(\d+)(:(\d+))?/ );
                        if ( p ) {
                            this.log.debug( 5, "%1 for %2 handling chunk %3", this, edge, p[0] );
                            if ( p[3] ) {
                                offset = sign * ( parseInt(p[1]) * 3600 + parseInt(p[2]) * 60 + parseInt(p[4]) );
                            } else {
                                offset = sign * ( parseInt(p[1]) * 3600 + parseInt(p[2]) * 60 );
                            }
                            rest = "";
                        } else {
                            // invalid offset
                            this.log.err( "%1 can't parse offset in edge %2 at %3", this, edge, rest );
                            break;
                        }
                    }
                }
                this.log.debug( 5, "%1 final offset for %2 is %3", this, edge, offset );
                moment += offset;
            }
            return moment;
        }
        this.log.err( "%1 can't parse schedule event %2", this, edge );
        return 0;
    }

    /** performOnEntity() - we don't have any actions, we're just a data mining
     *  tool (read only), so we don't override the superclass method and let it
     *  handle anything that comes up (which will be nothing).
     */
    /* performOnEntity( entity, actionName, parameters ) { } */

    /** Default action handler for sys_system.restart.
     */
    action_sys_system_restart( entity, params ) {
        this.stopTimer();
        this.startDelay( 1 );
        return Promise.resolve();
    }
};
